"""Manager code for spawned processes in multiprocessing implementation
of PDF extractor."""

import sys
import threading
import multiprocessing as mp
from multiprocessing.managers import BaseManager, BaseProxy

from typing import Optional, Generator

from .child import FitzWorker


class GeneratorProxy(BaseProxy):
    """Proxy type for generator objects."""

    _exposed_ = ['__next__']

    def __iter__(self):
        return self

    def __next__(self):
        return self._callmethod('__next__')


class WorkerProxy(BaseProxy):
    """Proxy type for FitzWorker methods.

    This code will run in the child processes, when triggered by
    the registered methods of the Manager.
    """

    filename: Optional[str] = None

    @classmethod
    def _open(cls, filename):
        cls.filename = filename

    @classmethod
    def _count_pages(cls) -> int:
        w = FitzWorker(cls.filename)
        return w.page_count()

    @classmethod
    def _list_pages(cls) -> Generator[str, None, None]:
        w = FitzWorker(cls.filename)
        return w.iter_contents()

    @classmethod
    def _extract_pages(cls, entries, save_path: str) -> Generator[str, None, None]:
        w = FitzWorker(cls.filename)
        for e in entries:
            w.extract_file(e, save_path)
            yield e


class FitzManager(BaseManager):
    """Multiprocessing manager to hold proxied worker callables."""

    pass


FitzManager.register('open', WorkerProxy._open)
FitzManager.register('page_count', WorkerProxy._count_pages)
FitzManager.register('iter_contents', WorkerProxy._list_pages, proxytype=GeneratorProxy)
FitzManager.register('extract_pages', WorkerProxy._extract_pages, proxytype=GeneratorProxy)


class FitzProcessWrangler(threading.local):
    """Thread-local state object holding a FitzManager instance.

    This is necessary so that each Mcomix extractor thread has its own
    FitzManager instance (and, therefore, its own worker process).
    """

    def __init__(self, filename, log_level):
        self.mgr = FitzManager()
        self.mgr.start()
        self.mgr.open(filename)
        self.log = mp.get_logger()
        if log_level is not None:
            self.log.setLevel(log_level)

    def page_count(self) -> int:
        """Get the number of pages in the PDF."""
        return self.mgr.page_count()

    def iter_contents(self) -> Generator[str, None, None]:
        """Return an iterator over all the page filenames in the PDF."""
        return self.mgr.iter_contents()

    def extract_pages(self, page_list, destination_dir) -> Generator[str, None, None]:
        """Extract the listed pages to the given directory."""
        return self.mgr.extract_pages(page_list, destination_dir)


# Test code, this module can be called directly with one argument (a PDF
# filename), and will print a list of all pages in the PDF (produced by
# a spawned FitzWorker process)
if __name__ == "__main__":
    mp.freeze_support()
    mp.set_start_method('spawn')
    infile = sys.argv[1]
    import logging
    wrangler = FitzProcessWrangler(infile, log_level=logging.DEBUG)

    print(f"All pages in PDF {infile}:")
    for file in wrangler.iter_contents():
        print(f"  {file}")



"""lens.py - Magnifying lens."""

import math

from gi.repository import Gdk, GdkPixbuf, Gtk

from mcomix.preferences import prefs
from mcomix import image_tools
from mcomix import constants
from mcomix import box
from mcomix import tools


class MagnifyingLens(object):

    """The MagnifyingLens creates cursors from the raw pixbufs containing
    the unscaled data for the currently displayed images. It does this by
    looking at the cursor position and calculating what image data to put
    in the "lens" cursor.

    Note: The mapping is highly dependent on the exact layout of the main
    window images, thus this module isn't really independent from the main
    module as it uses implementation details not in the interface.
    """

    def __init__(self, window):
        self._window = window
        self._area = self._window._main_layout
        self._area.connect('motion-notify-event', self._motion_event)

        #: Stores lens state
        self._enabled = False
        #: Stores a tuple of the last mouse coordinates
        self._point = None
        #: Stores the last rectangle that was used to render the lens
        self._last_lens_rect = None

    def get_enabled(self):
        return self._enabled

    def set_enabled(self, enabled):
        self._enabled = enabled

        if enabled:
            # FIXME: If no file is currently loaded, the cursor will still be hidden.
            self._window.cursor_handler.set_cursor_type(constants.NO_CURSOR)
            self._window.osd.clear()

            if self._point:
                self._draw_lens(*self._point)
        else:
            self._window.cursor_handler.set_cursor_type(constants.NORMAL_CURSOR)
            self._clear_lens()
            self._last_lens_rect = None

    enabled = property(get_enabled, set_enabled)

    def _draw_lens(self, x, y):
        """Calculate what image data to put in the lens and update the cursor
        with it; <x> and <y> are the positions of the cursor within the
        main window layout area.
        """
        if self._window.images[0].get_storage_type() not in (Gtk.ImageType.PIXBUF,
            Gtk.ImageType.ANIMATION):
            return

        lens_size = (prefs['lens size'],) * 2 # 2D only
        border_size = 1
        rectangle = self._calculate_lens_rect(x, y, *lens_size, border_size)

        draw_region = Gdk.Rectangle()
        draw_region.x, draw_region.y, draw_region.width, draw_region.height = rectangle
        if self._last_lens_rect:
            last_region = Gdk.Rectangle()
            last_region.x, last_region.y, last_region.width, last_region.height = self._last_lens_rect
            draw_region = Gdk.rectangle_union(draw_region, last_region)

        pixbuf = self._get_lens_pixbuf(x, y, lens_size, border_size,
            (x - rectangle[0], y - rectangle[1]))
        window = self._window._main_layout.get_bin_window()
        window.begin_paint_rect(draw_region)

        self._clear_lens()

        cr = window.cairo_create()
        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, 0, window)
        cr.set_source_surface(surface, rectangle[0], rectangle[1])
        cr.paint()

        window.end_paint()

        self._last_lens_rect = rectangle

    def _calculate_lens_rect(self, x, y, width, height, border_size):
        """ Calculates the area where the lens will be drawn on screen. This method takes
        screen space into calculation and moves the rectangle accordingly when the the rectangle
        would otherwise flow over the allocated area. """

        lens_x = max(x - width // 2, 0)
        lens_y = max(y - height // 2, 0)

        max_width, max_height = self._window.get_visible_area_size()
        max_width += int(self._window._hadjust.get_value())
        max_height += int(self._window._vadjust.get_value())
        lens_x = min(lens_x, max_width - width)
        lens_y = min(lens_y, max_height - height)

        return lens_x, lens_y, width + 2 * border_size, height + 2 * border_size

    def _clear_lens(self, current_lens_region=None):
        """ Invalidates the area that was damaged by the last call to draw_lens. """

        if not self._last_lens_rect:
            return

        window = self._window._main_layout.get_bin_window()
        crect = Gdk.Rectangle()
        crect.x, crect.y, crect.width, crect.height = self._last_lens_rect
        window.invalidate_rect(crect, True)
        window.process_updates(True)
        self._last_lens_rect = None

    def toggle(self, action):
        """Toggle on or off the lens depending on the state of <action>."""
        self.enabled = action.get_active()

    def _motion_event(self, widget, event):
        """ Called whenever the mouse moves over the image area. """
        self._point = (int(event.x), int(event.y))
        if self.enabled:
            self._draw_lens(*self._point)

    def _get_lens_pixbuf(self, x, y, lens_size, border_size, check_offset):
        """Get a pixbuf containing the appropiate image data for the lens
        where <x> and <y> are the positions of the cursor.
        """
        cb = self._window.layout.get_content_boxes()
        source_pixbufs = self._window.imagehandler.get_pixbufs(len(cb))
        transforms = self._window.transforms
        lens_scale = (prefs['lens magnification'],) * 2 # 2D only
        opaque = prefs['checkered bg for transparent images'] or not any(
            map(GdkPixbuf.Pixbuf.get_has_alpha, source_pixbufs))
        canvas = GdkPixbuf.Pixbuf.new(colorspace=GdkPixbuf.Colorspace.RGB,
            has_alpha=not opaque, bits_per_sample=8, width=lens_size[0],
            height=lens_size[1]) # 2D only
        canvas.fill(image_tools.convert_rgb16list_to_rgba8int(self._window.get_bg_colour()))
        for b, source_pixbuf, tf in zip(cb, source_pixbufs, transforms):
            if image_tools.is_animation(source_pixbuf):
                continue
            cpos = b.get_position()
            _scale, rotation, flips = tf.to_image_transforms() # FIXME use scale as soon as it is correctly included
            composite_color_args = image_tools.get_composite_color_args(0) if \
                source_pixbuf.get_has_alpha() and opaque else None
            self._draw_lens_pixbuf((x - cpos[0], y - cpos[1]), b.get_size(),
                source_pixbuf, rotation, flips,
                lens_size, lens_scale, canvas, prefs['scaling quality'],
                composite_color_args, (x - border_size - check_offset[0],
                y - border_size - check_offset[1])) # 2D only

        canvas = self._window.enhancer.enhance(canvas)

        return image_tools.add_border(canvas, border_size)

    def _draw_lens_pixbuf(self, ref_pos, csize, srcbuf, rotation, flips,
        lens_size, lens_scale, dstbuf, interpolation, composite_color_args,
        check_offset):
        if tools.volume(csize) == 0:
            return

        # Some computations are the same for each axis.
        def calc_1d(ref_pos, csize, src_pixbuf_size, lens_size, lens_scale):
            # compute initial scales, sizes and positions
            page_scale = csize / src_pixbuf_size
            source_ref_pos = ref_pos / page_scale
            combined_source_scale = page_scale * lens_scale
            mapped_ref_pos = source_ref_pos * combined_source_scale
            mapped_ref_pos_int = int(round(mapped_ref_pos * 2)) // 2
            mapped_size = int(round(src_pixbuf_size * combined_source_scale))
            # take rounding errors into account
            applied_source_scale = mapped_size / src_pixbuf_size
            # calculate data for clamping
            lens_size_2q, lens_size_2r = divmod(lens_size, 2)
            neg_mapped_lens_pos = lens_size_2q - mapped_ref_pos_int
            dest_lens_offset = neg_mapped_lens_pos
            dest_lens_end = dest_lens_offset + mapped_size
            # clamp to lens
            dest_lens_end = min(dest_lens_end, lens_size)
            dest_lens_offset = max(0, dest_lens_offset)
            dest_lens_size = dest_lens_end - dest_lens_offset
            return applied_source_scale, neg_mapped_lens_pos, dest_lens_offset, \
                dest_lens_size, mapped_size, mapped_ref_pos_int, lens_size_2q, lens_size_2r

        # prepare actual computation
        src_pixbuf_size = [srcbuf.get_width(), srcbuf.get_height()] # 2D only
        transpose = (1, 0) if tools.rotation_swaps_axes(rotation) else (0, 1) # 2D only
        tp = lambda x: tools.remap_axes(x, transpose)
        axis_flip = tuple(map(lambda r, f: (rotation in r) ^ f, ((270, 180), (90, 180)), tp(flips))) # 2D only

        # calculate size and position data
        applied_source_scale, neg_mapped_lens_pos, dest_lens_offset, dest_lens_size, \
            mapped_size, mapped_ref_pos_int, lens_size_2q, lens_size_2r = \
            [list(x) for x in zip(*(map(calc_1d, tp(ref_pos), tp(csize),
            src_pixbuf_size, tp(lens_size), tp(lens_scale))))]

        if min(dest_lens_size) > 0:
            # Using GdkPixbuf.Pixbuf.scale here so we do not need to worry about
            # interpolation issues when close to the edges. Also, one can exploit it
            # later to only recompute the parts of the lens where the content might
            # have changed.
            if any(flips) or any(axis_flip):
                # Unfortuantely, GdkPixbuf does not seem to provide an API for applying
                # arbitrary matrix transforms the same way, which is why we need to
                # apply inefficient workarounds.

                # keep track of (mirrored) reference point
                refpos_tracking = list(mapped_ref_pos_int)
                for i, s in enumerate(axis_flip):
                    if s:
                        # Subtracting the remainder keeps a lens with an odd number
                        # of pixels centered at the (mirrored) reference point.
                        refpos_tracking[i] = mapped_size[i] - refpos_tracking[i] - lens_size_2r[i]
                refpos_tracking = tools.vector_sub(refpos_tracking, lens_size_2q)

                # write to temporary buffer
                tempbuf = GdkPixbuf.Pixbuf.new(srcbuf.get_colorspace(),
                    srcbuf.get_has_alpha(), srcbuf.get_bits_per_sample(), *dest_lens_size)
                temp_lens_box = box.Box.intersect(box.Box(lens_size, position=refpos_tracking),
                    box.Box(mapped_size))
                srcbuf.scale(tempbuf, 0, 0, *dest_lens_size,
                    *tools.vector_opposite(temp_lens_box.get_position()),
                    *applied_source_scale, interpolation) # 2D only

                # apply all necessary transforms to temporary buffer
                tempbuf = image_tools.rotate_pixbuf(tempbuf, rotation)
                for i, f in enumerate(flips):
                    if f:
                        tempbuf = image_tools.flip_pixbuf(tempbuf, i)

                # Not sure whether it should be inverse axis remapping instead of
                # forward, but in 2D, there is no difference anyway.
                remapped_dest_lens_offset = tp(dest_lens_offset)
                remapped_dest_lens_size = tp(dest_lens_size)

                # copy result from temporary buffer to actual lens buffer
                if composite_color_args is None:
                    tempbuf.copy_area(0, 0, *remapped_dest_lens_size, dstbuf,
                        *remapped_dest_lens_offset) # 2D only
                else:
                    tempbuf.composite_color(dstbuf, *remapped_dest_lens_offset,
                        *remapped_dest_lens_size, *remapped_dest_lens_offset, 1, 1,
                        GdkPixbuf.InterpType.NEAREST, 255,
                        *tools.vector_add(tp(dest_lens_offset), check_offset),
                        *composite_color_args) # 2D only
                # unref temporary buffer
                tempbuf = None
            else:
                # no workaround needed
                if composite_color_args is None:
                    srcbuf.scale(dstbuf, *dest_lens_offset, *dest_lens_size,
                        *neg_mapped_lens_pos, *applied_source_scale, interpolation) # 2D only
                else:
                    srcbuf.composite_color(dstbuf, *dest_lens_offset, *dest_lens_size,
                        *neg_mapped_lens_pos, *applied_source_scale, interpolation,
                        255, *tools.vector_add(dest_lens_offset, check_offset),
                        *composite_color_args) # 2D only
        else:
            # If we are here, there is either no image to be drawn at all, or it is
            # out of range.
            pass
        return dstbuf


# vim: expandtab:sw=4:ts=4



# -*- coding: utf-8 -*-

''' MobiPocket handling (extract pictures) for MComix.

    Based on code from mobiunpack by Charles M. Hannum et al.
'''

import os
import re
import struct

from gi.repository import Gio

from mcomix import image_tools
from mcomix.archive import archive_base

class UnpackException(Exception):
    pass

class Sectionizer:
    def __init__(self, f):
        self.f = f
        header = self.f.read(78)
        self.ident = header[0x3C:0x3C+8]
        self.num_sections, = struct.unpack_from('>H', header, 76)
        sections = self.f.read(self.num_sections*8)
        self.sections = struct.unpack_from('>%dL' % (self.num_sections*2), sections, 0)[::2] + (0x7fffffff, )

    def loadSection(self, section, limit=0x7fffffff):
        before, after = self.sections[section:section+2]
        self.f.seek(before)
        if limit > after - before:
            limit = after - before
        return self.f.read(limit)

class MobiArchive(archive_base.NonUnicodeArchive):
    def __init__(self, archive):
        super(MobiArchive, self).__init__(archive)
        f = open(archive, 'rb')
        try:
            self.file = f
            self.sect = Sectionizer(self.file)
            if self.sect.ident != b'BOOKMOBI':
                raise UnpackException('invalid file format')
            self.header = self.sect.loadSection(0)
            self.crypto_type, = struct.unpack_from('>H', self.header, 0xC)
            if self.crypto_type != 0:
                raise UnpackException('file is encrypted')
            self.firstimg, = struct.unpack_from('>L', self.header, 0x6C)
        except:
            self.file = None
            f.close()
            raise

    def _close(self):
        if self.file is not None:
            self.file.close()
            self.file = None

    def iter_contents(self):
        ''' List archive contents. '''
        supported_mimes={}
        for mimes,exts in image_tools.get_supported_formats().values():
            ext = next(iter(exts))
            for mime in mimes:
                supported_mimes[mime] = ext
        for i in range(self.firstimg, self.sect.num_sections):
            magic = self.sect.loadSection(i, 10)
            mime, uncertain = Gio.content_type_guess(data=magic)
            mime = mime.lower()
            if mime in supported_mimes:
                ext = supported_mimes[mime]
                yield "image%05d.%s" % (1+i-self.firstimg, ext)

    def extract(self, filename, destination_dir):
        ''' Extract <filename> from the archive to <destination_dir>. '''
        destination_path = os.path.join(destination_dir, filename)
        fnparts = re.split(r'^image([0-9]*)\.', filename)
        if len(fnparts) == 3:
            i = int(fnparts[1])-1+self.firstimg
            data = self.sect.loadSection(i)
            with self._create_file(destination_path) as new:
                new.write(data)
        return destination_path

    def close(self):
        ''' Close the archive handle '''
        self._close()



"""Child process code for native PDF multiprocessing."""

import io
import os
import multiprocessing as mp
from PIL import Image
from typing import Generator, Optional

try:
    import fitz
except ImportError:
    import fitz_old as fitz

from mcomix.constants import PDF_RENDER_DPI_DEF
from mcomix.preferences import prefs


# Will delimit the page name from the xref part of a file name
XREF_DELIMITER = '_mcmxref'


class FitzWorker:
    def __init__(self, filename: Optional[str], log_level: Optional[int] = None) -> None:
        self._extension: Optional[str] = None
        self._complex_doc = False
        self.log = mp.get_logger()
        if log_level is not None:
            self.log.setLevel(log_level)
        self.doc = fitz.open(filename)

    def page_count(self) -> int:
        return self.doc.page_count

    def _image_extension(self, page_num: int) -> str:
        """Return the filename extension for the page image."""
        if self._extension is None:
            self._extension = self._check_image_type(page_num)
        return self._extension

    def _extract_as_image(self, page_num: int) -> bool:
        """Check whether the page can be extracted via image export."""
        if self._complex_doc:
            return False
        if not self._must_render_page(page_num) and self._can_extract_image(page_num):
            return True
        return False

    def _must_render_page(self, page_num: int) -> bool:
        """Determine if a page has any forced-render markers.

        Rendering to pixmap must be forced if any of these apply:
            - The page has any text content
            - The page has any drawing content
            - The page contains 0, or more than 1, embedded image
        """
        page = self.doc[page_num]
        if len(page.get_images()) != 1 or len(page.get_text()) > 0:
            self._complex_doc = True
            result = True
            self.log.debug("PDF page %d, must render page", page_num + 1)
        else:
            result = False
            self.log.debug("PDF page %d, rendering not forced", page_num + 1)
        del page
        return result

    def _can_extract_image(self, page_num: int) -> bool:
        """Determine if a page has an extractable image.

        Makes a closer examination than _must_render_page(),
        by checking whether the page contains a single, full-page
        image, then actually extracting the first such image
        encountered to determine the filetype (extension).

        (Subsequent embedded images are assumed to have the same
        type as the first full-page image encountered. This may
        be somewhat fragile, but it's a huge performance boost.)
        """
        page = self.doc[page_num]
        image_info = page.get_image_info()
        if len(image_info) != 1:
            self.log.debug(
                "PDF page %d, cannot extract. Image count = %d",
                page_num + 1, len(image_info))
            return False
        info = image_info[0]
        img_rect = fitz.Rect(info.get('bbox', (0, 0, 0, 0))).irect
        page_rect = fitz.Rect(page.mediabox).irect
        page_area = page_rect.get_area()
        area_diff = abs(page_area - img_rect.get_area())
        is_full_page: bool = area_diff < 0.05 * page_area
        if is_full_page:
            self.log.debug('PDF page %d: can extract fullpage image', page_num + 1)
        else:
            self.log.debug(
                'PDF page %d, cannot extract image: %s',
                page_num + 1, f"img_rect={img_rect}, page_rect={page_rect}")
        del page
        del image_info
        return is_full_page

    def _check_image_type(self, page_num: int) -> str:
        """Examine the page's embedded image for its file type.

        The extension is determined heuristically by probing only the
        first page of the document for an embedded image, then using
        its type.

        If the first page does have a single embedded image, it's
        assumed that _all_ pages contain an image of the same type.
        This may be a fragile assumption.

        If the probe fails, 'png' is used as a fallback, as that's the
        type rendered page pixmaps will be saved with.
        """
        extension = 'png'
        try:
            xref = self._get_image_xref(page_num)
            img = fitz.image_profile(
                self.doc.xref_stream_raw(xref))
            # If image_profile returns an empty dict, the image type is
            # "exotic" and not supported for direct extraction.
            # That doesn't mean that the images can't be extracted.
            # Document.extract_image will automatically convert to PNG,
            # when we call it to extract the xref. It'll be slower than
            # extraction without converting, but still very fast.
            if img:
                extension = img.get('ext', 'png')
                del img
        except (AttributeError, TypeError):
            pass
        finally:
            return extension

    def _get_image_xref(self, page_num: int) -> int:
        try:
            image_info = self.doc.get_page_images(page_num)
            xref = int(image_info[0][0])
            return xref
        except (TypeError, IndexError):
            return -1
        finally:
            image_info = None

    def iter_contents(self) -> Generator[str, None, None]:
        for pg in range(self.doc.page_count):
            pagenum = f"page{pg + 1:04}"
            if self._extract_as_image(pg):
                xref = self._get_image_xref(pg)
                ext = self._image_extension(pg)
                filename = f"{pagenum}{XREF_DELIMITER}{xref:04}.{ext}"
            else:
                filename = f"{pagenum}.png"
            yield filename

    def extract_xref(self, page: int, xref: int, path: str) -> None:
        """Save the embedded PDF image for a given xref. The page is indexed starting with zero."""
        img = self.doc.extract_image(xref)
        if not img:
            return
        os.makedirs(os.path.dirname(path), exist_ok=True)
        img_bytes = img.get("image", b"")
        del img

        # The extract_image method always returns the unrotated version of the image,
        # unaffected by any page modifications of rotation.
        rotation = self.doc[page].rotation
        if rotation in (90, 180, 270) and prefs['auto rotate from exif']:
            buffer = io.BytesIO(img_bytes)
            pil_img = Image.open(buffer)
            transpose = Image.Transpose.ROTATE_270
            if rotation == 180:
                transpose = Image.Transpose.ROTATE_180
            elif rotation == 270:
                transpose = Image.Transpose.ROTATE_90
            pil_img = pil_img.transpose(transpose)
            pil_img.save(path)

        else:
            with open(path, "wb") as out:
                out.write(img_bytes)
        del img_bytes

    def render_page(self, pg: int, path: str) -> None:
        """Render the page to an image file and save."""
        page = self.doc[pg]
        pixmap = page.get_pixmap(dpi=PDF_RENDER_DPI_DEF)
        pixmap.save(path)
        del pixmap
        del page

    def extract_file(self, filename: str, dest: str) -> None:
        outpath = os.path.join(dest, filename)
        if XREF_DELIMITER in filename:
            pginfo, ref = filename.split(XREF_DELIMITER)
            page = int(pginfo[-4:]) - 1
            xref = int(ref[:4])
            self.extract_xref(page, xref, outpath)
        elif filename.startswith('page'):
            pg_num = int(filename[4:8]) - 1
            self.render_page(pg_num, outpath)



"""labels.py - Gtk.Label convenience classes."""

from mcomix import i18n
from gi.repository import GLib, Gtk, Pango


class FormattedLabel(Gtk.Label):
    """FormattedLabel keeps a label always formatted with some pango weight,
    style and scale, even when new text is set using set_text().
    """

    _STYLES = {
        Pango.Style.NORMAL: 'normal',
        Pango.Style.OBLIQUE: 'oblique',
        Pango.Style.ITALIC: 'italic',
    }

    def __init__(self, text: str = '', weight: Pango.Weight = Pango.Weight.NORMAL,
                 style: Pango.Style = Pango.Style.NORMAL, scale: float = 1.0) -> None:
        super(FormattedLabel, self).__init__()
        self._weight = weight
        self._style = style
        self._scale = scale
        self.set_text(text)

    def set_text(self, text: str) -> None:
        markup = '<span font_size="%u" font_weight="%u" font_style="%s">%s</span>' % (
            int(self._scale * 10 * 1024),
            self._weight,
            self._STYLES[self._style],
            GLib.markup_escape_text(i18n.to_display_string(text))
        )
        self.set_markup(markup)


class BoldLabel(FormattedLabel):
    """A FormattedLabel that is always bold and otherwise normal."""

    def __init__(self, text: str = '') -> None:
        super(BoldLabel, self).__init__(text=text, weight=Pango.Weight.BOLD)


class ItalicLabel(FormattedLabel):
    """A FormattedLabel that is always italic and otherwise normal."""

    def __init__(self, text: str = '') -> None:
        super(ItalicLabel, self).__init__(text=text, style=Pango.Style.ITALIC)


# vim: expandtab:sw=4:ts=4



# -*- coding: utf-8 -*-

"""Multiprocessing PDF handler."""

import multiprocessing as mp
from mcomix.archive import archive_base
from mcomix.archive.native_pdf.manager import FitzProcessWrangler
from mcomix.log import getLevel

from typing import Generator, List


class FitzArchive(archive_base.BaseArchive):
    """PDF file reader/extractor using PyMuPDF."""

    # Concurrent calls to extract welcome!
    support_concurrent_extractions = True

    def __init__(self, archive) -> None:
        """Initialize the object, first as a BaseArchive."""
        super().__init__(archive)
        self.log = mp.get_logger()
        self.log.setLevel(getLevel())
        self.log.debug("PDF contains %d pages", self.mgr.page_count())

    @staticmethod
    def is_available() -> bool:
        """Report whether this extractor is available (always true)."""
        return True

    def _open_doc(self) -> FitzProcessWrangler:
        """Create a new FitzManager instance for processes accessing the archive."""
        self.close()
        self._mgr = FitzProcessWrangler(self.archive, log_level=self.log.level)
        return self._mgr

    def close(self) -> None:
        """Destroy the wrangler object and free resources."""
        if hasattr(self, '_mgr'):
            del self._mgr

    @property
    def mgr(self) -> FitzProcessWrangler:
        if not hasattr(self, '_mgr'):
            return self._open_doc()
        return self._mgr

    def iter_contents(self) -> Generator[str, None, None]:
        """Generate page filenames."""
        return self.mgr.iter_contents()

    def extract(self, filename, destination_dir):
        """Extract the named page file to a directory."""
        self._create_directory(destination_dir)
        output = list(self.mgr.extract_pages([filename], destination_dir))
        if len(output) > 0:
            return True
        return False

    def iter_extract(self, entries: List[str], destination_dir: str) -> Generator[str, None, None]:
        """Return a generator of extracted filepaths."""
        self._create_directory(destination_dir)
        return self.mgr.extract_pages(entries, destination_dir)



"""MComix - GTK Comic Book Viewer
"""

# -------------------------------------------------------------------------
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# -------------------------------------------------------------------------

import multiprocessing as mp
from .run import run


def main() -> None:
    run()


if __name__ == '__main__':
    mp.freeze_support()
    mp.set_start_method('spawn')
    main()



# -*- coding: utf-8 -*-

"""Shim module to conditionally load the FitzArchive class."""

import os
from typing import Type

from mcomix import log
from mcomix.archive.archive_base import BaseArchive, DisabledArchive

from mcomix.version_tools import LegacyVersion

FITZ_VERSION_REQUIRED = "1.19.2"


class DisabledError(RuntimeError): pass

class UnsupportedFitzVersionError(ImportError):
    def __init__(
            self, message=None, name=None, path=None,
            found_version=None):
        self.found_version = found_version
        self.minimum_version = FITZ_VERSION_REQUIRED
        if message is None:
            message = f"PyMuPDF {self.minimum_version} or later required"
            if self.found_version is not None:
                message += f", {self.found_version} found"
        super().__init__(message, name=name, path=path)


class DisabledFitzArchive(DisabledArchive):
    """Subclass of DisabledArchive used when FitzArchive is unavailable.

    This class will masquerade as FitzArchive for purposes of upstream
    reporting, so that the correct class is logged as being unavailable."""

    __name__ = "FitzArchive"


# On import, this code tests for a compatible version of the fitz
# module (PyMuPDF), and exports as "PdfMultiArchive" either the
# FitzArchive class from native_pdf.parent, or the DisabledFitzArchive
# class (aliased to 'FitzArchive' for caller-side reporting purposes)

try:
    if os.environ.get("MCOMIX_DISABLE_PDF_MULTI") is not None:
        raise DisabledError("MCOMIX_DISABLE_PDF_MULTI set in environment")
    try:
        import fitz
    except ImportError:
        import fitz_old as fitz

    fitz_version = LegacyVersion(fitz.VersionFitz)
    required_version = LegacyVersion(FITZ_VERSION_REQUIRED)

    if fitz_version < required_version:
        raise UnsupportedFitzVersionError(found_version=fitz.VersionFitz)
    from mcomix.archive.native_pdf.parent import FitzArchive

    log.info("Native PDF handler loaded, PyMuPDF version %s", fitz.VersionFitz)
    PdfMultiArchive: Type[BaseArchive] = FitzArchive
except (DisabledError, ImportError) as ex:
    log.info("Can't enable pdf_multi: %s", str(ex))
    PdfMultiArchive = DisabledFitzArchive

# vim: expandtab:sw=4:ts=4



"""Portability functions for MComix."""

import ctypes
import locale
import sys

from mcomix import constants


def uri_prefix() -> str:
    """The prefix used for creating file URIs. This is 'file://' on
    Linux, but 'file:' on Windows due to urllib using a different
    URI creating scheme here."""
    if sys.platform == "win32":
        return "file:"
    else:
        return "file://"


def normalize_uri(uri: str) -> str:
    """Normalize URIs passed into the program by different applications,
    normally via drag-and-drop."""

    if uri.startswith("file://localhost/"):  # Correctly formatted.
        return uri[16:]
    elif uri.startswith("file:///"):  # Nautilus etc.
        return uri[7:]
    elif uri.startswith("file:/"):  # Xffm etc.
        return uri[5:]
    else:
        return uri


def invalid_filesystem_chars() -> str:
    """List of characters that cannot be used in filenames on the target platform."""
    if sys.platform == "win32":
        return r':*?"<>|' + "".join([chr(i) for i in range(0, 32)])
    else:
        return ""


def get_default_locale() -> str:
    """Gets the user's default locale."""
    if sys.platform == "win32":
        windll = ctypes.windll.kernel32
        code = windll.GetUserDefaultUILanguage()
        return locale.windows_locale[code]
    else:
        lang, _ = locale.getdefaultlocale(("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG"))
        if lang:
            return str(lang)
        else:
            return "C"


def is_system_ui_dark_themed() -> constants.SystemThemeLightness:
    """Determine if the system is configured to use a dark theme by default."""
    if sys.platform == "win32":
        import winreg

        try:
            with winreg.OpenKey(
                winreg.HKEY_CURRENT_USER,
                "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Themes\\Personalize",
            ) as personalize_handle:
                _, values_count, _ = winreg.QueryInfoKey(personalize_handle)
                for key_index in range(values_count):
                    key_name, key_value, _ = winreg.EnumValue(
                        personalize_handle, key_index
                    )

                    if key_name == "AppsUseLightTheme":
                        if key_value == 0:
                            return constants.SystemThemeLightness.DARK
                        else:
                            return constants.SystemThemeLightness.LIGHT

                return constants.SystemThemeLightness.LIGHT
        except OSError:
            return constants.SystemThemeLightness.UNKNOWN
    else:
        return constants.SystemThemeLightness.UNKNOWN


# vim: expandtab:sw=4:ts=4



"""tools.py - Contains various helper functions."""

import bisect
import gc
import itertools
import math
import operator
import os
import re
import sys
from functools import reduce
from typing import (Any, Iterable, List, Mapping, Sequence, Tuple, TypeVar,
                    Union)

Numeric = TypeVar('Numeric', int, float)

NUMERIC_REGEXP = re.compile(r"\d+|\D+")  # Split into numerics and characters
PREFIXED_BYTE_UNITS = ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB", "RiB", "QiB")


def cmp(a: str, b: str) -> int:
    """ Forward port of Python2's cmp function """
    return (a > b) - (a < b)


class AlphanumericSortKey:
    """ Compares two strings by their natural order (i.e. 1 before 10) """
    def __init__(self, filename: str) -> None:
        self.filename_parts: List[Union[int, str]] = [
            int(part) if part.isdigit() else part
            for part in NUMERIC_REGEXP.findall(filename.lower())
        ]

    def __lt__(self, other: 'AlphanumericSortKey') -> bool:
        for left, right in itertools.zip_longest(self.filename_parts, other.filename_parts, fillvalue=''):
            if not isinstance(left, type(right)):
                left_str = str(left)
                right_str = str(right)
                if left_str < right_str:
                    return True
                elif left_str > right_str:
                    return False
            else:
                if left < right:
                    return True
                elif left > right:
                    return False

        return False


def alphanumeric_sort(filenames: List[str]) -> None:
    """Do an in-place alphanumeric sort of the strings in <filenames>,
    such that for an example "1.jpg", "2.jpg", "10.jpg" is a sorted
    ordering.
    """

    filenames.sort(key=AlphanumericSortKey)


def bin_search(lst: List, value: Any) -> int:
    """ Binary search for sorted list C{lst}, looking for C{value}.
    @return: List index on success. On failure, it returns the 1's
    complement of the index where C{value} would be inserted.
    This implies that the return value is non-negative if and only if
    C{value} is contained in C{lst}. """

    index = bisect.bisect_left(lst, value)
    if index != len(lst) and lst[index] == value:
        return index
    else:
        return ~index


def get_home_directory() -> str:
    """On UNIX-like systems, this method will return the path of the home
    directory, e.g. /home/username. On Windows, it will return an MComix
    sub-directory of <Documents and Settings/Username>.
    """
    if sys.platform == 'win32':
        return os.path.join(os.path.expanduser('~'), 'MComix')
    else:
        return os.path.expanduser('~')


def get_config_directory() -> str:
    """Return the path to the MComix config directory. On UNIX, this will
    be $XDG_CONFIG_HOME/mcomix, on Windows it will be in %APPDATA%/MComix.

    See http://standards.freedesktop.org/basedir-spec/latest/ for more
    information on the $XDG_CONFIG_HOME environmental variable.
    """
    if sys.platform == 'win32':
        return os.path.join(os.path.expandvars('%APPDATA%'), 'MComix')
    else:
        base_path = os.getenv('XDG_CONFIG_HOME',
                              os.path.join(get_home_directory(), '.config'))
        return os.path.join(base_path, 'mcomix')


def get_data_directory() -> str:
    """Return the path to the MComix data directory. On UNIX, this will
    be $XDG_DATA_HOME/mcomix, on Windows it will be the same directory as
    get_config_directory().

    See http://standards.freedesktop.org/basedir-spec/latest/ for more
    information on the $XDG_DATA_HOME environmental variable.
    """
    if sys.platform == 'win32':
        return os.path.join(os.path.expandvars('%APPDATA%'), 'MComix')
    else:
        base_path = os.getenv('XDG_DATA_HOME',
                              os.path.join(get_home_directory(), '.local/share'))
        return os.path.join(base_path, 'mcomix')


def number_of_digits(n: int) -> int:
    if 0 == n:
        return 1
    return int(math.log10(abs(n))) + 1


def decompose_byte_size_exponent(n: float) -> Tuple[float, int]:
    e = 0
    while n > 1024.0:
        n /= 1024.0
        e += 1
    return (n, e)


def byte_size_exponent_to_prefix(e: int) -> str:
    return PREFIXED_BYTE_UNITS[min(e, len(PREFIXED_BYTE_UNITS) - 1)]


def format_byte_size(n: float) -> str:
    nn, e = decompose_byte_size_exponent(n)
    return ('%d %s' if nn == int(nn) else '%.1f %s') % \
        (nn, byte_size_exponent_to_prefix(e))


def garbage_collect() -> None:
    """ Runs the garbage collector. """
    gc.collect(0)


def div(a: Numeric, b: Numeric) -> float:
    return float(a) / float(b)


def volume(t: List[int]) -> int:
    return reduce(operator.mul, t, 1)


def relerr(approx: Numeric, ideal: Numeric) -> float:
    return abs(div(approx - ideal, ideal))


def smaller(a: List, b: List) -> List:
    """ Returns a list with the i-th element set to True if and only if the i-th
    element in a is less than the i-th element in b. """
    return list(map(operator.lt, a, b))


def smaller_or_equal(a: List, b: List) -> List:
    """ Returns a list with the i-th element set to True if and only if the i-th
    element in a is less than or equal to the i-th element in b. """
    return list(map(operator.le, a, b))


def scale(t: Sequence[Numeric], factor: Numeric) -> List[Numeric]:
    return [x * factor for x in t]


def vector_sub(a: List[Numeric], b: List[Numeric]) -> List[Numeric]:
    """ Subtracts vector b from vector a. """
    return list(map(operator.sub, a, b))


def vector_add(a: List[Numeric], b: List[Numeric]) -> List[Numeric]:
    """ Adds vector a to vector b. """
    return list(map(operator.add, a, b))


def vector_opposite(a: List[Numeric]) -> List[Numeric]:
    """ Returns the opposite vector -a. """
    return list(map(operator.neg, a))


def remap_axes(vector, order):
    return [vector[i] for i in order]


def inverse_axis_map(order):
    identity = list(range(len(order)))
    return [identity[order[i]] for i in identity]


def compile_rotations(*rotations):
    return reduce(lambda a, x: a + (x % 360) % 360, rotations, 0)


def rotation_swaps_axes(rotation: int) -> bool:
    return rotation in (90, 270)


def fixed_strings_regex(strings: Iterable[str]) -> str:
    # introduces a matching group
    unique_strings = set(strings)
    return r'(%s)' % '|'.join(sorted([re.escape(s) for s in unique_strings]))


def formats_to_regex(formats: Mapping) -> re.Pattern:
    """ Returns a compiled regular expression that can be used to search for
    file extensions specified in C{formats}. """
    return re.compile(r'\.' + fixed_strings_regex(
        itertools.chain.from_iterable([e[1] for e in formats.values()])) + r'$', re.I)


def append_number_to_filename(filename: str, number: int) -> str:
    """ Generate a new string from filename with an appended number right
    before the extension. """
    file_no_ext = os.path.splitext(filename)[0]
    ext = os.path.splitext(filename)[1]
    return file_no_ext + (" (%s)" % (number)) + ext

# vim: expandtab:sw=4:ts=4



""" A series of a few simple linear transforms. """

# Allow class name in annotations while still defining class
from __future__ import annotations

import sys
from typing import Optional, Tuple


class Matrix:
    """Simple linear transformations represented as a 2x2 matrix.

    Note that the current implementation does not support arbitrary
    matrices. Instead, the predefined Matrix objects and conversion
    functions in the Transform class should be used to create Matrix
    objects."""

    def __init__(self, m1: float, m2: float, m3: float, m4: float) -> None:
        """Initialize a row-major transformation matrix."""
        self.m: Tuple[float, float, float, float] = (m1, m2, m3, m4)

    def __str__(self) -> str:
        return str(self.m)

    def __repr__(self) -> str:
        args = ", ".join([str(x) for x in self.m])
        return f'Matrix({args})'

    def __bool__(self) -> bool:
        """Does this matrix perform any operations (is not identity)?"""
        return self != Transform.ID

    def __add__(self, other: object) -> Matrix:
        """Combine Matrices: self + other."""
        if not isinstance(other, Matrix):
            return NotImplemented
        a = self.m
        b = other.m
        return Matrix(
            b[0] * a[0] + b[1] * a[2],
            b[0] * a[1] + b[1] * a[3],
            b[2] * a[0] + b[3] * a[2],
            b[2] * a[1] + b[3] * a[3],
        )

    def __radd__(self, other: object) -> Matrix:
        """Combine Matrices: other + self."""
        if isinstance(other, Matrix):
            return other + self
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        """Test equality: self == other."""
        if isinstance(other, Matrix):
            return len(self.m) == len(other.m) and all([
                self.m[i] == other.m[i]
                for i in range(len(self.m))
            ])
        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Test inequality: self != other."""
        if isinstance(other, Matrix):
            return not self == other
        return NotImplemented

    def and_then(self, next: Matrix) -> Matrix:
        """The matrix result of transforming self by 'next'."""
        return self + next

    def and_then_all(self, *nexts: Matrix) -> Matrix:
        """The matrix resulting from self transformed by each 'nexts' in turn."""
        out = self
        for m in nexts:
            out += m
        return out

    def swaps_axes(self) -> bool:
        """ Determines whether this matrix includes a transpose of the axes. """
        return self.m[0] == 0

    def rotated(self, deg: int) -> Matrix:
        """The matrix result of rotating self by 'deg' degrees."""
        return self + Transform.from_rotation(deg)

    def scaled(self, s0: float, s1: Optional[float] = None) -> Matrix:
        """The matrix result of scaling self by the provided factor(s).

        t.scale(s) will scale both dimensions by s.
        t.scale(s0, s1) will scale x by s0, and y by s1."""
        if s1 is None:
            s1 = s0
        return self + Transform.from_scales(s0, s1)

    def flipped(self, x: bool = False, y: bool = False) -> Matrix:
        """The matrix result of applying the specified flips."""
        if x and y:
            return self + Transform.ROT180
        if x:
            return self + Transform.INVX
        if y:
            return self + Transform.INVY
        return self

    def flipped_x(self) -> Matrix:
        """The matrix result of flipping self horizontally."""
        return self.flipped(x=True)

    def flipped_y(self) -> Matrix:
        """The matrix result of flipping self vertically."""
        return self.flipped(y=True)

    def to_image_transforms(self) -> Tuple[Tuple[float, float], int, Tuple[bool, bool]]:
        """ Decomposes the transform to a sequence of hints to basic transform
        instructions typically found in image processing libraries. The sequence
        will refer to the positive scaling factors to be applied for each axis
        first, followed by at most one rotation, and finally followed by at most
        one flip.
        @return a tuple (s, r, f) where s is a sequence of positive scaling
        factors for the corresponding axes, r is one of (0, 90, 180, 270),
        referring to the clockwise rotation to be applied, and f is a sequence
        of bools where True refers to the corresponding axis to be flipped. """
        s: Tuple[float, float] = (abs(self.m[0] + self.m[1]), abs(self.m[2] + self.m[3]))
        r: int = 90 if self.swaps_axes() else 0
        f: Tuple[bool, bool] = (
            self.swaps_axes() ^ (self.m[0] < 0 or self.m[1] < 0),
            (self.m[2] < 0 or self.m[3] < 0)
        )
        if all(f):
            f = (False, False)
            r += 180
        if f[0] and r == 90:
            # Try to avoid horizontal flips because they tend to be slow, given
            # a certain combination of hardware, memory layout and libraries.
            f = (False, True)
            r = 270
        return (s, r, f)


class Transform(Matrix):

    ID = Matrix(1, 0, 0, 1)
    ROT90 = Matrix(0, -1, 1, 0)
    ROT180 = Matrix(-1, 0, 0, -1)
    ROT270 = Matrix(0, 1, -1, 0)
    INVX = Matrix(-1, 0, 0, 1)
    INVY = Matrix(1, 0, 0, -1)
    TRP = Matrix(0, 1, 1, 0)
    TRPINV = Matrix(0, -1, -1, 0)

    @classmethod
    def __call__(cls, *values: float) -> Matrix:
        """Construct a new Matrix by calling Transform(values)."""
        return Matrix(*values)

    @classmethod
    def from_rotation(cls, deg: int) -> Matrix:
        """Get the predefined Matrix for a supported rotation."""
        if abs(deg) not in (0, 90, 180, 270):
            raise ValueError("illegal rotation angle: " + str(deg))
        if deg < 0: deg = 360 + deg
        if deg == 90:
            return cls.ROT90
        if deg == 180:
            return cls.ROT180
        if deg == 270:
            return cls.ROT270
        return cls.ID

    @classmethod
    def from_scales(cls, s0: float, s1: float) -> Matrix:
        """Get a Matrix representing the given x and y scales."""
        if s0 == 0 or s1 == 0:
            raise ValueError("illegal scaling factor(s): " + str((s0, s1)))
        return Matrix(s0, 0, 0, s1)

    @classmethod
    def from_flips(cls, x: bool, y: bool) -> Matrix:
        """Get the predefined Matrix representing the given x and y flips."""
        if x and y:
            return cls.ROT180
        if x:
            return cls.INVX
        if y:
            return cls.INVY
        return cls.ID

    @classmethod
    def from_image_transforms(
            cls,
            t: Tuple[Tuple[float, float], int, Tuple[bool, bool]]
    ) -> Matrix:
        """Create a Matrix transform from a set of image transforms."""
        s, r, f = t[0:3]
        return (
            cls.from_scales(*s)
            + cls.from_rotation(r)
            + cls.from_flips(*f)
        )

# vim: expandtab:sw=4:ts=4
